import asyncio
import json
import logging
import os
import re
import subprocess
from base64 import b64decode
from functools import cached_property
from importlib.metadata import version
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any

import boto3
from pydantic import BaseModel, validator

from sagemaker_shim.logging import STDOUT_LEVEL
from sagemaker_shim.utils import safe_extract

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
else:
    S3Client = object


logger = logging.getLogger(__name__)

BUCKET_NAME_REGEX = re.compile(r"^[a-zA-Z0-9.\-_]{1,255}$")
BUCKET_ARN_REGEX = re.compile(
    r"^arn:(aws).*:(s3|s3-object-lambda):[a-z\-0-9]*:[0-9]{12}:accesspoint[/:][a-zA-Z0-9\-.]{1,63}$|^arn:(aws).*:s3-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:][a-zA-Z0-9\-]{1,63}[/:]accesspoint[/:][a-zA-Z0-9\-]{1,63}$"
)


class InferenceIO(BaseModel):
    relative_path: Path
    bucket_name: str
    bucket_key: str
    decompress: bool = False

    class Config:
        frozen = True

    # TODO validation of attrs

    def local_file(self, path: Path) -> Path:
        # TODO handle images and non relative_paths
        return path / self.relative_path

    def download(self, *, input_path: Path, s3_client: S3Client) -> None:
        dest_file = self.local_file(path=input_path)

        logger.info(
            f"Downloading {self.bucket_key=} from {self.bucket_name=} to {dest_file=}"
        )

        dest_file.parent.mkdir(exist_ok=True, parents=True)

        if self.decompress:
            with TemporaryDirectory() as tmp_dir:
                # TODO add tests
                zipfile = Path(tmp_dir) / "src.zip"
                with zipfile.open("wb") as f:
                    s3_client.download_fileobj(
                        Bucket=self.bucket_name,
                        Key=self.bucket_key,
                        Fileobj=f,
                    )
                safe_extract(src=zipfile, dest=dest_file.parent)
        else:
            with dest_file.open("wb") as f:
                s3_client.download_fileobj(
                    Bucket=self.bucket_name,
                    Key=self.bucket_key,
                    Fileobj=f,
                )

    def upload(self, *, output_path: Path, s3_client: S3Client) -> None:
        src_file = str(self.local_file(path=output_path))

        logger.info(
            f"Uploading {src_file=} to {self.bucket_name=} with {self.bucket_key=}"
        )

        # TODO limit file size
        s3_client.upload_file(
            Filename=src_file,
            Bucket=self.bucket_name,
            Key=self.bucket_key,
        )


class InferenceResult(BaseModel):
    return_code: int
    outputs: list[InferenceIO]
    sagemaker_shim_version: str = version("sagemaker-shim")

    class Config:
        frozen = True


class InferenceTask(BaseModel):
    pk: str
    inputs: list[InferenceIO]
    output_bucket_name: str
    output_prefix: str

    class Config:
        frozen = True

    @validator("output_prefix")
    def validate_prefix(cls, v: str) -> str:  # noqa:B902
        if not v:
            raise ValueError("Prefix cannot be blank")

        if v[-1] != "/":
            v += "/"

        return v

    @validator("output_bucket_name")
    def validate_bucket_name(cls, v: str) -> str:  # noqa:B902
        if BUCKET_NAME_REGEX.match(v) or BUCKET_ARN_REGEX.match(v):
            return v
        else:
            raise ValueError("Invalid bucket name")

    @staticmethod
    def decode_b64j(*, encoded: str | None) -> Any:
        if encoded is None:
            return None
        else:
            return json.loads(
                b64decode(encoded.encode("utf-8")).decode("utf-8")
            )

    @classmethod
    @property
    def cmd(cls) -> Any:
        cmd = cls.decode_b64j(
            encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_CMD_B64J")
        )
        logger.debug(f"{cmd=}")
        return cmd

    @classmethod
    @property
    def entrypoint(cls) -> Any:
        entrypoint = cls.decode_b64j(
            encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J")
        )
        logger.debug(f"{entrypoint=}")
        return entrypoint

    @classmethod
    @property
    def input_path(cls) -> Path:
        input_path = Path(
            os.environ.get("GRAND_CHALLENGE_COMPONENT_INPUT_PATH", "/input")
        )
        logger.debug(f"{input_path=}")
        return input_path

    @classmethod
    @property
    def output_path(cls) -> Path:
        output_path = Path(
            os.environ.get("GRAND_CHALLENGE_COMPONENT_OUTPUT_PATH", "/output")
        )
        logger.debug(f"{output_path=}")
        return output_path

    @cached_property
    def _s3_client(self) -> S3Client:
        return boto3.client(
            "s3", endpoint_url=os.environ.get("AWS_S3_ENDPOINT_URL")
        )

    @property
    def proc_args(self) -> list[str]:
        """
        Implementation of CMD and ENTRYPOINT parsing from
        https://docs.docker.com/engine/reference/builder/#understand-how-cmd-and-entrypoint-interact
                                    | No ENTRYPOINT              | ENTRYPOINT exec_entry p1_entry | ENTRYPOINT ["exec_entry", "p1_entry"]
        No CMD                      | error, not allowed         | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry
        CMD ["exec_cmd", "p1_cmd"]  | exec_cmd p1_cmd            | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry exec_cmd p1_cmd
        CMD ["p1_cmd", "p2_cmd"]    | p1_cmd p2_cmd              | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry p1_cmd p2_cmd
        CMD exec_cmd p1_cmd         | /bin/sh -c exec_cmd p1_cmd | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry /bin/sh -c exec_cmd p1_cmd
        """  # noqa: B950
        entrypoint = self.entrypoint
        cmd = self.cmd

        if entrypoint is None and cmd is None:
            raise ValueError("Either cmd or entrypoint must be set")
        elif isinstance(entrypoint, str):
            proc_args = ["/bin/sh", "-c", entrypoint]
        else:
            proc_args = entrypoint if entrypoint is not None else []

            if cmd is None:
                pass
            elif isinstance(cmd, str):
                proc_args += ["/bin/sh", "-c", cmd]
            else:
                proc_args += cmd

        return proc_args

    @property
    def proc_env(self) -> dict[str, str]:
        """The environment for the subprocess"""
        env = os.environ.copy()

        # Set LD_LIBRARY_PATH correctly, see
        # https://pyinstaller.org/en/stable/runtime-information.html#ld-library-path-libpath-considerations
        lp_key = "LD_LIBRARY_PATH"
        lp_orig = env.get(lp_key + "_ORIG")

        if lp_orig is not None:
            env[lp_key] = lp_orig
        else:
            env.pop(lp_key, None)

        return env

    async def invoke(self) -> InferenceResult:
        # TODO ensure only one algorithm is running
        # TODO ensure /input is empty
        outputs = set()

        self.download_input()

        return_code = await self.execute()

        if return_code == 0:
            outputs = self.upload_output()

        # TODO get algorithm runtime
        # TODO cleanup /input and /output

        return InferenceResult(return_code=return_code, outputs=outputs)

    def download_input(self) -> None:
        for input_file in self.inputs:
            input_file.download(
                input_path=self.input_path, s3_client=self._s3_client
            )

    def upload_output(self) -> set[InferenceIO]:
        output_path = self.output_path
        outputs: set[InferenceIO] = set()

        for f in output_path.rglob("**/*"):
            if not f.is_file() and not f.is_symlink():
                logger.warning(f"Skipping {f=}")
            else:
                relative_path = f.relative_to(output_path)
                output = InferenceIO(
                    relative_path=relative_path,
                    bucket_key=f"{self.output_prefix}{relative_path}",
                    bucket_name=self.output_bucket_name,
                )
                output.upload(
                    output_path=self.output_path, s3_client=self._s3_client
                )
                outputs.add(output)

        return outputs

    async def execute(self) -> int:
        logger.debug(f"Calling {self.proc_args=}")

        process = await asyncio.create_subprocess_exec(
            *self.proc_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self.proc_env,
        )

        try:
            await asyncio.gather(
                self._stream_to_external(
                    stream=process.stdout, level=STDOUT_LEVEL
                ),
                self._stream_to_external(
                    stream=process.stderr, level=STDOUT_LEVEL + 10
                ),
            )
        except Exception:
            process.kill()
            raise
        finally:
            return_code = await process.wait()

        logger.debug(f"{return_code=}")

        return return_code

    async def _stream_to_external(
        self, *, stream: asyncio.StreamReader | None, level: int
    ) -> None:
        # TODO find a way to test the logging
        if stream is None:
            return

        while True:
            line = await stream.readline()

            if not line:
                break

            self.log_external(
                level=level,
                msg=line.replace(b"\x00", b"").decode("utf-8"),
            )

    def log_external(self, *, level: int, msg: str) -> None:
        logger.log(
            level=level, msg=msg, extra={"internal": False, "task": self}
        )



import asyncio
import logging
import logging.config
import sys
from functools import wraps
from typing import Any, Callable

import click
import uvicorn

from sagemaker_shim.app import app
from sagemaker_shim.logging import LOGGING_CONFIG

logger = logging.getLogger(__name__)


def cli_coroutine(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group()
def cli() -> None:
    pass


@cli.command(short_help="Start the model server")
def serve() -> None:
    logging.config.dictConfig(LOGGING_CONFIG)
    uvicorn.run(app=app, host="0.0.0.0", port=8080, log_config=None)


if __name__ == "__main__":
    # https://pyinstaller.org/en/stable/runtime-information.html#run-time-information
    we_are_bundled = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")

    if we_are_bundled:
        # https://pyinstaller.org/en/stable/runtime-information.html#using-sys-executable-and-sys-argv-0
        cli(sys.argv[1:])
    else:
        cli()



import logging
import re
import zipfile
from os.path import commonpath
from pathlib import Path

from sagemaker_shim.vendor.werkzeug.security import safe_join

logger = logging.getLogger(__name__)


def _filter_members(members: list[zipfile.ZipInfo]) -> list[dict[str, str]]:
    """Filter common prefixes and uninteresting files from a zip archive"""
    members = [
        m
        for m in members
        if not m.is_dir()
        and re.search(r"(__MACOSX|\.DS_Store|desktop.ini)", m.filename) is None
    ]

    # Remove any common parent directories
    if len(members) == 1:
        path = str(Path(members[0].filename).parent)
        path = "" if path == "." else path
    else:
        path = commonpath([m.filename for m in members])

    if path:
        sliced_path = slice(len(path) + 1, None, None)
    else:
        sliced_path = slice(None, None, None)

    return [
        {"src": m.filename, "dest": m.filename[sliced_path]} for m in members
    ]


def safe_extract(*, src: Path, dest: Path) -> None:
    """
    Safely extracts a zip file into a directory

    Any common prefixes and system files are removed.
    """

    if not dest.exists():
        raise RuntimeError("The destination must exist")

    with src.open("rb") as f:
        with zipfile.ZipFile(f) as zf:
            members = _filter_members(zf.infolist())

            for member in members:
                file_dest = safe_join(str(dest), member["dest"])

                if file_dest is None:
                    raise RuntimeError("Suspicious file operation")

                # We know that the dest is within the prefix as
                # safe_join is used, and the destination is already
                # created, so ok to create the parents here
                file_dest.parent.mkdir(exist_ok=True, parents=True)

                logger.info(
                    f"Extracting {member['src']=} from {src} to {file_dest}"
                )

                with zf.open(member["src"], "r") as fs, open(
                    file_dest, "wb"
                ) as fd:
                    while True:
                        chunk = fs.read(1024)
                        if not chunk:
                            break

                        fd.write(chunk)



import json
import logging
import os
from typing import Any

logger = logging.getLogger(__name__)

STDOUT_LEVEL = logging.INFO


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record=record)

        # We need to explicitly add the source as an annotation
        # as CloudWatch unifies the logs
        if record.levelno <= STDOUT_LEVEL:
            source = "stdout"
        else:
            source = "stderr"

        internal = getattr(record, "internal", True)
        task = getattr(record, "task", None)
        task_pk = getattr(task, "pk", None)

        # TODO find a way to test the logging

        return "\n".join(
            json.dumps(
                {
                    "log": m,
                    "level": record.levelname,
                    "source": source,
                    "internal": internal,
                    "task": task_pk,
                }
            )
            for m in message.splitlines()
        )


class StdStreamFilter(logging.Filter):
    def __init__(self, *args: Any, stdout: bool, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__stdout = stdout

    def filter(self, record: logging.LogRecord) -> bool:
        """Should this log message be displayed?"""
        if self.__stdout:
            # stdout, STDOUT_LEVEL and lower
            return record.levelno <= STDOUT_LEVEL
        else:
            # stderr, greater than STDOUT_LEVEL
            return record.levelno > STDOUT_LEVEL


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "stdout": {
            "()": StdStreamFilter,
            "stdout": True,
        },
        "stderr": {
            "()": StdStreamFilter,
            "stdout": False,
        },
    },
    "formatters": {
        "json": {
            "()": JSONFormatter,
        },
    },
    "handlers": {
        "console_stdout": {
            "class": "logging.StreamHandler",
            "formatter": "json",
            "stream": "ext://sys.stdout",
            "filters": ["stdout"],
        },
        "console_stderr": {
            "class": "logging.StreamHandler",
            "formatter": "json",
            "stream": "ext://sys.stderr",
            "filters": ["stderr"],
        },
    },
    "root": {
        "level": os.environ.get("LOGLEVEL", "INFO").upper(),
        "handlers": ["console_stdout", "console_stderr"],
    },
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sagemaker_shim', 'sagemaker_shim.vendor', 'sagemaker_shim.vendor.werkzeug']

package_data = \
{'': ['*']}

install_requires = \
['boto3', 'click', 'fastapi', 'uvicorn[standard]']

entry_points = \
{'console_scripts': ['sagemaker-shim = sagemaker_shim.cli:cli']}

setup_kwargs = {
    'name': 'sagemaker-shim',
    'version': '0.0.8',
    'description': 'Adapts algorithms that implement the Grand Challenge inference API for running in SageMaker',
    'long_description': '# SageMaker Shim for Grand Challenge\n\n[![CI](https://github.com/jmsmkn/sagemaker-shim/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/jmsmkn/sagemaker-shim/actions/workflows/ci.yml?query=branch%3Amain)\n[![PyPI](https://img.shields.io/pypi/v/sagemaker-shim)](https://pypi.org/project/sagemaker-shim/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/sagemaker-shim)](https://pypi.org/project/sagemaker-shim/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nThis repo contains a library that adapts algorithms that implement the Grand Challenge inference API for running in SageMaker.\n\nThe application contains:\n\n- A `click` cli client with options to launch a web server\n- A `fastapi` web server that implements the SageMaker endpoints\n- and `pydantic` models that interface between S3, and run the original inference jobs.\n\nThe application is compiled on Python 3.10 using `pyinstaller`, and then distributed as a statically linked binary using `staticx`.\nIt is able to adapt any container, including ones based on `scratch` or `alpine` images.\n\n## Usage\n\nThe binary is designed to be added to an existing container image that implements the Grand Challenge API.\nOn Grand Challenge this happens automatically by using [crane](https://github.com/google/go-containerregistry/blob/main/cmd/crane/doc/crane_mutate.md) to add the binary, directories and environment variables to each comtainer image.\nThe binary itself will:\n\n1. Download the input files from the provided locations on S3 to `/input`, optionally decompressing the inputs.\n1. Execute the original container program in a subprocess.\n   This is found by inspecting the following environment variables:\n    - `GRAND_CHALLENGE_COMPONENT_CMD_B64J`: the original `cmd` of the container, json encoded as a base64 string.\n    - `GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J`: the original `entrypoint` of the container, json encoded as a base64 string.\n1. Upload the contents of `/output` to the given output S3 bucket and prefix.\n\n### `sagemaker-shim serve`\n\nThis starts the webserver on http://0.0.0.0:8080 which implements the [SageMaker API](https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-batch-code.html).\nThere are three endpoints:\n\n- `/ping` (GET): returns an empty 200 response if the container is healthy\n- `/execution-parameters` (GET): returns the preferred execution parameters for AWS SageMaker Batch Inference\n- `/invocations` (POST): SageMaker can make POST requests to this endpoint.\n  The body contains the json encoded data required to run a single inference task:\n\n  ```json\n    {\n        "pk": "unique-test-id",\n        "inputs": [\n            {\n                "relative_path": "interface/path",\n                "bucket_name": "name-of-input-bucket",\n                "bucket_key": "/path/to/input/file/in/bucket",\n                "decompress": false,\n            },\n            ...\n        ],\n        "output_bucket_name": "name-of-output-bucket",\n        "output_prefix": "/prefix/of/output/files",\n    }\n  ```\n\n  The endpoint will return an object containing the return code of the subprocess in `response["return_code"]`,\n  and any outputs will be placed in the output bucket at the output prefix.\n\n### Patching an Existing Container\n\nTo patch an existing container image in a registry see the example in [tests/utils.py](tests/utils.py).\nFirst you will need to get the original `cmd` and `entrypoint` using `get_new_env_vars` and `get_image_config`.\nThen you can add the binary, set the new `cmd`, `entrypoint`, and environment variables with `mutate_image`.\n',
    'author': 'James Meakin',
    'author_email': '12661555+jmsmkn@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DIAGNijmegen/rse-sagemaker-shim',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
